# -*- coding:utf-8 -*-
import os, csv, sys, math
import numpy as np
from scipy.fftpack import dst, idst
from scipy import interpolate, signal
from numba import jit, f8, i8, b1 # comment out this line and all @jit lines if numba does not exist
# physical constants
epsilon0 = 8.85418782 * (10.0 ** -12.0); e = 1.60217657 * (10.0 ** -19.0);
me = 9.10938291 * (10.0 ** -31.0); mi = 1836.0 * me; kB = 1.3806488 * (10.0 ** -23.0)
# plasma conditions
n0 = 1e7; ne = n0; Te = (100.0 * e/kB); vd = 400.0e3; Esw = mi * vd * vd / 2.0;
ld = np.sqrt(epsilon0*kB*Te/(e**2 * ne)); kd = 1.0/ld
V0_arr = [4.0e3]; rw_arr = [0.15] # set V0 and r_w
@jit
def calc_one_column(data, Wn, Wm, W, dx, N, debye):
    for j in range(N):
        denom = debye + 4.0 - Wm - 1.0/Wm - Wn - 1.0/Wn
        if(denom != 0.0): data[j] = data[j] * dx * dx / denom
        Wn = Wn * W
@jit
def solve_on_kspace(rho_k, N, dx, debye):
    Wm = 1.0; Wn = 1.0; W = np.exp(2.0*np.pi*1.0j/N)
    for i in range(N):
        calc_one_column(rho_k[i], Wn, Wm, W, dx, N, debye)
        Wm = Wm * W
    return rho_k
@jit
def solve_poisson_with_debye(rho, N, dx, debye):
    phi_x = np.real(np.fft.ifft2(
        solve_on_kspace(np.fft.fft2(rho, s=(int(N),int(N))), N, dx, debye),
        s=(int(N),int(N))))
    return phi_x
def get_cmat(rw, x, y, N, dx, debye):
    div = int(rw/dx) + 1 # grid number within tether's radius
    # Count c-matrix index
    cmat_num = []
    for i in range(N/2 - div, N/2 + div + 1):
        xv2 = x[i] * x[i]
        for j in range(N/2 - div, N/2 + div + 1):
            r = np.sqrt(xv2 + y[j]**2)
            if(r <= rw): cmat_num.append([i, j])

    icmat = len(cmat_num)
    rho = np.zeros(shape=(len(x), len(y)), dtype="float64")
    cmat = np.zeros(shape=(icmat, icmat))
    # ITERATE: Add unit charge on one c-matrix and Solve poisson.
    for (l, cm) in enumerate(cmat_num):
        print "next cmat: %d/%d." % (l+1, icmat)
        i = cm[0]; j = cm[1]; rho[i][j] = 1.0/epsilon0
        phi_x = solve_poisson_with_debye(rho, N, dx, debye)
        for (m, cm2) in enumerate(cmat_num):
            cmat[m][l] = phi_x[ cm2[0] ][ cm2[1] ]
        rho[i][j] = 0.0
    # Then obtain C = B'^-1
    cmat = np.linalg.inv(cmat); return cmat, cmat_num
@jit
def reallocate_rho(rho, V0, phi, cmat, cmat_num, dx):
    # Reallocate d_rho to make the tether equipotential
    drho = np.zeros(shape=(len(cmat_num)))
    for (l, cm) in enumerate(cmat_num):
        for (m, cm2) in enumerate(cmat_num):
            ii = cm2[0]; jj = cm2[1]; drho[l] += cmat[l][m] * (V0 - phi[ii][jj])

    for (l, cm) in enumerate(cmat_num):
        i = cm[0]; j = cm[1]; rho[i][j] += drho[l]/epsilon0
# search the position at which eV(r) - Esw get minimum
def search_solution(kinetic_energy, phi_x, N):
    minval = 100.0; mini = 0;
    if phi_x.max() < kinetic_energy:
        print "max phi = %f < kinetic_energy %f" % (phi_x.max(), kinetic_energy); return -1
    else:
        # search only in front of tether
        for i in range(0, N/2+1):
            val = abs(kinetic_energy - phi_x[i])
            if abs(minval) > abs(val): minval = val; mini = i
        print "Minimal diff: Esw - phi = %f" % minval
        return mini
@jit
def Ueff(x, phi_x, N, p, l):
    r = p/l; minval = 100.0; mini = 0; res = 100.0 # calculation variables
    # search l_max only in front of the tether, x < 0.0
    for i in range(0, N/2+1):
        val = -x[i] - r
        # get the position at which x[i] - p/l gets minimum.
        if abs(minval) > abs(val): minval = val; mini = i
    res = (l**2) + e * phi_x[mini] / Esw
    # Ueff must be lower than 1.0
    while(res > 1.0):
        mini = mini - 1 # move the index - 1
        res = (l**2) + e * phi_x[mini] / Esw
    return res
@jit
def calcLmax(x, phi_x, N, p):
    minval = 100.0; mini = 0 # calculation variables
    # search l_max only in front of the tether, x < 0.0
    for i in range(0, N/2+1):
        l = p/(-x[i]); val = (l**2) + e * phi_x[i]/Esw - 1.0
        if val < 0.0 and abs(minval) > abs(val): minval = val; mini = i

    # linear interpolation
    # Ueff < 1 | Ueff == 1.0 | Ueff > 1
    # x[mini]  | true l_max  | x[mini+1]
    # -----------------------------> x-axis
    next_val = (p/-x[mini+1])**2 + e*phi_x[mini+1]/Esw - 1.0
    local_dx = -x[mini] + x[mini+1]

    dval_dx = (next_val - minval)/local_dx
    true_x = -x[mini] + (-minval/dval_dx)
    return p/true_x
@jit
def calcChi(x, phi_x, N, p, dl):
    l_min = p/(2.0*ld); l_max = calcLmax(x, phi_x, N, p)
    delta_one = 0.0; l = l_min; local_dl = dl
    for i in range(0, int( (l_max - l_min) / dl )+1):
        val = 1.0 - Ueff(x, phi_x, N, p, l)
        delta_one += local_dl/np.sqrt(val)
        # if l + dl > l_max (one step before the end of calc.), set l = l_max
        if(l_max - l < dl):
            local_dl = l_max - l
            l = l_max
        else: l = l + dl

    delta_two = np.sin(l_min)
    chi = np.pi - 2.0 * (delta_one + delta_two)
    return chi
if __name__ == '__main__':
    N = 8192; h = 0.03; max_x = h*N; max_y = h*N
    x = np.arange(-max_x/2.0, max_x/2.0, h); y = np.arange(-max_y/2.0, max_y/2.0, h) # generate grids

    Pdyn = mi * n0 * (vd**2); debye = (h**2)*(kd**2); kinetic_energy = mi * (vd**2) / (2.0 * e)
    for rw in rw_arr:
        rho = np.zeros(shape=(len(x), len(y)), dtype="float64")
        # calculate capancitance matrix
        cmat, cmat_num = get_cmat(rw, x, y, N, h, debye)

        for V0 in V0_arr:
            rho.fill(0.0)
            phi = np.zeros(shape=(len(x), len(y)), dtype="float64")
            reallocate_rho(rho, V0, phi, cmat, cmat_num, h) # set rho along with cmatrix to obtain an equipotential
            phi = solve_poisson_with_debye(rho, N, h, debye)
            phi = phi[N/2] # extract 1d profile

            # search the solution of eV(r) == Esw
            i = search_solution(kinetic_energy, phi, N)
            thrust = 0.0; rs = 0.0
            if i >= 0 and i < N:
                rs = abs(x[i])
                # calculate the thrust by IonScattering Method
                dp = 0.003; dl = 0.0015 # integration parameters
                for j in range(1, int(rs/dp)+1):
                    chi_per_two = calcChi(x, phi, N, j*dp, dl) / 2.0
                    thrust = thrust + dp * (np.sin(chi_per_two)**2)
                thrust = thrust * 4.0 * Pdyn * 1e9
            else:
                # minimum thrust (impinging momentum)
                rs = 2 * rw; thrust = Pdyn * 1e9 * 2.0 * rw
            print "Thrust [nN/m] = %f (Method 1) at %f " % (rs * 3.09 * Pdyn * 1e9, V0)
            print "Thrust [nN/m] = %f (Method 2) at %f " % (thrust, V0)
